/*
**  GraphQL-Tools-Sequelize -- Resolver Functions for GraphQL-Tools using Sequelize ORM
**  Copyright (c) 2016 Ralf S. Engelschall <rse@engelschall.com>
**
**  Permission is hereby granted, free of charge, to any person obtaining
**  a copy of this software and associated documentation files (the
**  "Software"), to deal in the Software without restriction, including
**  without limitation the rights to use, copy, modify, merge, publish,
**  distribute, sublicense, and/or sell copies of the Software, and to
**  permit persons to whom the Software is furnished to do so, subject to
**  the following conditions:
**
**  The above copyright notice and this permission notice shall be included
**  in all copies or substantial portions of the Software.
**
**  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
**  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
**  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
**  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
**  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
**  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
**  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

import UUID              from "pure-uuid"
import capitalize        from "capitalize"
import co                from "co"
import Ducky             from "ducky"
import Promise           from "bluebird"
import GraphQLFields     from "graphql-fields"

export default class GraphQLToolsSequelize {
    constructor (sequelize, authorizers = null, tracer = null) {
        this.sequelize   = sequelize
        this.models      = sequelize.models
        this.authorizers = authorizers
        this.tracer      = tracer
    }

    /*   optionally check authorization  */
    authorized (op, type, obj, ctx) {
        if (this.authorizers === null)
            return Promise.resolve(true)
        if (this.authorizers[type] === undefined)
            return Promise.resolve(true)
        let result
        try { result = this.authorizers[type](op, obj, ctx) }
        catch (ex) { result = Promise.resolve(false) }
        if (typeof result === "boolean")
            result = Promise.resolve(result)
        return result
    }

    /*   optionally give tracer a hint  */
    trace (oid, type, op, via, onto) {
        if (this.tracer === null)
            return Promise.resolve()
        let result = this.tracer.call(null, oid, type, op, via, onto)
        if (!(typeof result === "object" && typeof result.then === "function"))
            result = Promise.resolve(result)
        return result
    }

    /*  map scalar types between GraphQL, JavaScript and Sequelize  */
    mapScalarType (type, from, to) {
        const scalarTypeMap = [
            { graphql: "Boolean", javascript: "boolean", sequelize: "BOOLEAN" },
            { graphql: "Int",     javascript: "number",  sequelize: "INTEGER" },
            { graphql: "Float",   javascript: "number",  sequelize: "FLOAT" },
            { graphql: "String",  javascript: "string",  sequelize: "STRING" },
            { graphql: "ID",      javascript: "string",  sequelize: "STRING" }
        ]
        let found = scalarTypeMap.find((item) => item[from] === type)
        if (found !== undefined)
            found = found[to]
        return found
    }

    /*  determine fields (and their type) of a GraphQL object type  */
    fieldsOfGraphQLType (info, entity) {
        let fields = { attribute: {}, relation: {} }
        let fieldsAll = info.schema._typeMap[entity]._fields
        Object.keys(fieldsAll).forEach((field) => {
            let type = fieldsAll[field].type
            while (typeof type.ofType === "object")
                type = type.ofType
            if (type.constructor.name === "GraphQLScalarType")
                fields.attribute[field] = type.name
            else if (type.constructor.name === "GraphQLObjectType")
                fields.relation[field] = type.name
            else
                throw new Error(`unknown type "${type.constructor.name}" for field "${field}"`)
        })
        return fields
    }

    /*  determine fields (and their type) of a GraphQL request  */
    fieldsOfGraphQLRequest (args, info, entity) {
        let defined = this.fieldsOfGraphQLType(info, entity)
        let fields = { attribute: {}, relation: {} }
        Object.keys(args.with).forEach((name) => {
            if (defined.relation[name]) {
                let value = args.with[name]
                if (typeof value === "string")
                    value = { set: value }
                if (typeof value !== "object")
                    throw new Error(`invalid value for relation "${name}" on type "${entity}"`)
                if (typeof value.set === "string")
                    value.set = [ value.set ]
                if (typeof value.add === "string")
                    value.add = [ value.add ]
                if (typeof value.del === "string")
                    value.del = [ value.del ]
                if (!Ducky.validate(value, `{
                    set?: [ string* ], add?: [ string+ ], del?: [ string+ ] }`))
                    throw new Error(`invalid value for relation "${name}" on type "${entity}"`)
                fields.relation[name] = value
            }
            else if (defined.attribute[name]) {
                let type = this.mapScalarType(defined.attribute[name], "graphql", "javascript")
                if (typeof args.with[name] !== type)
                    throw new Error(`value for attribute "${name}" on type "${entity}" ` +
                        `has to be compatible with GraphQL type "${defined.attribute[name]}"`)
                fields.attribute[name] = args.with[name]
            }
            else
                throw new Error(`field "${name}" not known on type "${entity}"`)
        })
        return fields
    }

    /*  GraphQL standard options to Sequelize findById() options conversion  */
    findOneOptions (entity, args, info) {
        let opts = {}

        /*  determine allowed fields  */
        let allowed = this.fieldsOfGraphQLType(info, entity)

        /*  determine Sequelize "where" parameter  */
        if (args.where !== undefined) {
            if (typeof args.where !== "object")
                throw new Error(`invalid "where" argument`)
            opts.where = args.where
            opts.where = {}
            Object.keys(args.where).forEach((field) => {
                if (!allowed.attribute[field])
                    throw new Error(`invalid "where" argument: ` +
                        `no such field "${field}" on type "${entity}"`)
                opts.where[field] = args.where[field]
            })
        }

        /*  determine Sequelize "attributes" parameter  */
        let fieldInfo = GraphQLFields(info)
        let fields = Object.keys(fieldInfo)
        let attr = fields.filter((field) => allowed.attribute[field])
        let rels = fields.filter((field) => allowed.relation[field])
        if (rels.length === 0) {
            /*  in case no relationships should be followed at all from this entity,
                we can load the requested attributes only. If any relationship
                should be followed from this entity, we have to avoid
                such an attribute filter as this means that at least "hasOne" relationships
                would be "null" when dereferenced afterwards.  */
            if (attr.length === 0)
                /*  should not happen as GraphQL does not allow an entirely empty selection  */
                opts.attributes = [ this.sequelize.literal("1") ]
            else
                opts.attributes = attr
        }

        return opts
    }

    /*  GraphQL standard options to Sequelize findAll() options conversion  */
    findAllOptions (entity, args, info) {
        let opts = {}

        /*  determine allowed fields  */
        let allowed = this.fieldsOfGraphQLType(info, entity)

        /*  determine Sequelize "where" parameter  */
        if (args.where !== undefined) {
            if (typeof args.where !== "object")
                throw new Error(`invalid "where" argument`)
            opts.where = args.where
            opts.where = {}
            Object.keys(args.where).forEach((field) => {
                if (!allowed.attribute[field])
                    throw new Error(`invalid "where" argument: ` +
                        `no such field "${field}" on type "${entity}"`)
                opts.where[field] = args.where[field]
            })
        }

        /*  determine Sequelize "offset" parameter  */
        if (args.offset !== undefined)
            opts.offset = args.offset

        /*  determine Sequelize "limit" parameter  */
        if (args.limit !== undefined)
            opts.limit = args.limit

        /*  determine Sequelize "order" parameter  */
        if (args.order  !== undefined) {
            if (!Ducky.validate(args.order, `( string | [ (string | [ string, string ])+ ])`))
                throw new Error(`invalid "order" argument: wrong structure`)
            opts.order = args.order
        }

        /*  determine Sequelize "attributes" parameter  */
        let fieldInfo = GraphQLFields(info)
        let fields = Object.keys(fieldInfo)
        let attr = fields.filter((field) => allowed.attribute[field])
        let rels = fields.filter((field) => allowed.relation[field])
        if (rels.length === 0) {
            /*  in case no relationships should be followed at all from this entity,
                we can load the requested attributes only. If any relationship
                should be followed from this entity, we have to avoid
                such an attribute filter as this means that at least "hasOne" relationships
                would be "null" when dereferenced afterwards.  */
            if (attr.length === 0)
                /*  should not happen as GraphQL does not allow an entirely empty selection  */
                opts.attributes = [ this.sequelize.literal("1") ]
            else
                opts.attributes = attr
        }

        return opts
    }

    /*  API: generate GraphQL schema definition for entity query  */
    schemaEntityQuery (type) {
        return `` +
            `#  Query one **${type}** entity by its unique id.\n` +
            `${type}(id: String!): ${type}\n` +
            `#  Query all **${type}** entities by a condition (\`where\`), optionally sort them (\`order\`)\n` +
            `#  and optionally reduce the resulting entity list to a maximum number of entities (\`limit\`),\n` +
            `#  optionally starting the result set at the n-th entity (zero-based \`offset\`).\n` +
            `${type}s(where: JSON, order: JSON, offset: Int = 0, limit: Int = 100): [${type}]!\n`
    }

    /*  API: generate GraphQL schema definition for relationship query  */
    schemaRelationQueryOne (name, type) {
        return `` +
            `#  Query one **${type}** entity by following the **${name}** relationship.\n` +
            `#  The entity can be optionally filtered by a condition (\`where\`).\n` +
            `${name}(where: JSON): ${type}\n`
    }

    /*  API: generate GraphQL schema definition for relationship query  */
    schemaRelationQueryMany (name, type) {
        return `` +
            `#  Query one or many **${type}** entities by following the **${name}** relationship.\n` +
            `#  The entities can be filtered by a condition (\`where\`), optionally be sorted (\`order\`)\n` +
            `#  and optionally reduced to a maximum number of entities (\`limit\`),\n` +
            `#  optionally starting the result set at the n-th entity (zero-based \`offset\`).\n` +
            `${name}(where: JSON, offset: Int = 0, limit: Int = 100, order: JSON): [${type}]!\n`
    }

    /*  API: generate GraphQL schema definition for entity mutation  */
    schemaEntityMutation (type) {
        return `` +
            `# Create one **${type}** entity, optionally with specified attributes (\`with\`)\n` +
            `create${type}(with: JSON): ${type}!\n` +
            `# Clone one **${type}** entity (found by unique id), by cloning its attributes (but not its relationships).\n` +
            `clone${type}(id: String!): ${type}!\n` +
            `# Update one **${type}** entity (found by unique id) with specified attributes.\n` +
            `update${type}(id: String!, with: JSON!): ${type}!\n` +
            `# Update zero or more **${type}** entities (found by condition) with specified attributes.\n` +
            `update${type}s(where: JSON!, with: JSON!): [${type}]!\n` +
            `# Delete one **${type}** entity (found by unique id).\n` +
            `delete${type}(id: String! ): String\n` +
            `# Delete zero or more **${type}** entities (found by condition).\n` +
            `delete${type}s(where: JSON!): [String]!\n`
    }

    /*  API: direct query single entity by id  */
    queryEntityOne (type) {
        return co.wrap(function * (parent, args, ctx, info) {
            /*  find entity  */
            let opts = this.findOneOptions(type, args, info)
            let obj = yield (this.models[type].findById(args.id, opts))
            if (obj === null)
                return null

            /*  check authorization  */
            if (!(yield (this.authorized("read", type, obj, ctx))))
                return null

            /*  trace access  */
            yield (this.trace(obj.id, type, "read", "direct", "one"))

            return obj
        }.bind(this))
    }

    /*  API: direct query all entities by where  */
    queryEntityAll (type) {
        return co.wrap(function * (parent, args, ctx, info) {
            /*  find entities  */
            let opts = this.findAllOptions(type, args, info)
            let objs = yield(this.models[type].findAll(opts))

            /*  check authorizations  */
            objs = yield (Promise.filter(objs, (obj) => {
                return this.authorized("read", type, obj, ctx)
            }))

            /*  trace access  */
            yield (Promise.each(objs, (obj) => {
                return this.trace(obj.id, type, "read", "direct", "all")
            }))

            return objs
        }.bind(this))
    }

    /*  API: direct query single entity by 1-ary relationship  */
    queryRelationOne (type, getter) {
        return co.wrap(function * (parent, args, ctx, info) {
            /*  find entity  */
            let opts = this.findOneOptions(type, args, info)
            let obj = yield(parent[getter](opts))
            if (obj === null)
                return null

            /*  check authorization  */
            if (!(yield (this.authorized("read", type, obj, ctx))))
                return null

            /*  trace access  */
            yield (this.trace(obj.id, type, "read", "relation", "all"))

            return obj
        }.bind(this))
    }

    /*  API: direct query all entities by N-ary relationship  */
    queryRelationMany (type, getter) {
        return co.wrap(function * (parent, args, ctx, info) {
            /*  find entities  */
            let opts = this.findAllOptions(type, args, info)
            let objs = yield (parent[getter](opts))

            /*  check authorization  */
            objs = yield (Promise.filter(objs, (obj) => {
                return this.authorized("read", type, obj, ctx)
            }))

            /*  trace access  */
            yield (Promise.each(objs, (obj) => {
                return this.trace(obj.id, type, "read", "relation", "all")
            }))

            return objs
        }.bind(this))
    }

    /*  update all relation fields of an entity  */
    mutationEntityUpdateFields (type, obj, def, upd) {
        return co(function * () {
            let rels = Object.keys(upd)
            for (let i = 0; i < rels.length; i++) {
                let name  = rels[i]
                let value = upd[name]
                const changeRelation = co.wrap(function * (prefix, ids) {
                    for (let j = 0; j < ids.length; j++) {
                        let id   = ids[j]
                        let type = def[name]
                        let foreign = yield (this.models[type].findById(id))
                        if (foreign === null)
                            throw new Error(`no such entity ${type}#${id} found`)
                        let method = `${prefix}${capitalize(name)}`
                        if (typeof obj[method] === "function")
                            yield (obj[method](foreign))
                        else if (prefix === "add" || prefix === "remove") {
                            /*  special case for 1-arity relationship!  */
                            method = `set${capitalize(name)}`
                            if (typeof obj[method] === "function")
                                yield (obj[method](prefix === "add" ? foreign : null))
                            else
                                throw new Error(`relationship mutation method not found ` +
                                    `to ${prefix} relation ${name} on type ${type}`)
                        }
                    }
                }.bind(this))
                if (value.set)
                    yield (changeRelation("set",    value.set))
                if (value.del)
                    yield (changeRelation("remove", value.del))
                if (value.add)
                    yield (changeRelation("add",    value.add))
            }
        }.bind(this))
    }

    /*  API: direct create a single entity  */
    mutationEntityCreate (type) {
        return co.wrap(function * (parent, args, ctx, info) {
            /*  determine fields of entity as defined in GraphQL schema  */
            let defined = this.fieldsOfGraphQLType(info, type)

            /*  determine fields of entity as requested in GraphQL request  */
            let build = this.fieldsOfGraphQLRequest(args, info, type)

            /*  handle unique id  */
            if (build.attribute.id === undefined)
                /*  auto-generate the id  */
                build.attribute.id = (new UUID(1)).format()
            else {
                /*  ensure the id is unique  */
                let existing = yield (this.models[type].findById(build.attribute.id))
                if (existing !== null)
                    throw new Error(`entity ${type}#${build.attribute.id} already exists`)
            }

            /*  build a new entity  */
            let obj = this.models[type].build(build.attribute)

            /*  check access to entity  */
            if (!(yield (this.authorized("create", type, obj, ctx))))
                throw new Error(`not allowed to create entity of type "${type}"`)

            /*  save new entity  */
            let err = yield (obj.save().catch((err) => err))
            if (typeof err === "object" && err instanceof Error)
                throw new Error("Sequelize: save: " + err.message + ":" +
                    err.errors.map((e) => e.message).join("; "))

            /*  post-adjust the relationships according to the request  */
            yield (this.mutationEntityUpdateFields(type, obj,
                defined.relation, build.relation))

            /*  check access to entity again  */
            if (!(yield (this.authorized("read", type, obj, ctx))))
                return null

            /*  trace access  */
            yield (this.trace(obj.id, type, "create", "direct", "one"))

            /*  return new entity  */
            return obj
        }.bind(this))
    }

    /*  API: direct clone a single entity (without relationships)  */
    mutationEntityClone (type) {
        return co.wrap(function * (parent, args, ctx, info) {
            /*  determine fields of entity as defined in GraphQL schema  */
            let defined = this.fieldsOfGraphQLType(info, type)

            /*  load source entity  */
            let src = yield (this.models[type].findById(args.id))
            if (src === null)
                throw new Error(`source entity ${type}#${args.id} not found`)

            /*  check access to source entity  */
            if (!(yield (this.authorized("read", type, src, ctx))))
                throw new Error(`not allowed to read entity of type "${type}"`)

            /*  build a new entity  */
            let data = {}
            data.id = (new UUID(1)).format()
            Object.keys(defined.attribute).forEach((attr) => {
                if (attr !== "id")
                    data[attr] = src[attr]
            })
            let obj = this.models[type].build(data)

            /*  check access to entity  */
            if (!(yield (this.authorized("create", type, obj, ctx))))
                throw new Error(`not allowed to create entity of type "${type}"`)

            /*  save new entity  */
            let err = yield (obj.save().catch((err) => err))
            if (typeof err === "object" && err instanceof Error)
                throw new Error("Sequelize: save: " + err.message + ":" +
                    err.errors.map((e) => e.message).join("; "))

            /*  check access to entity again  */
            if (!(yield (this.authorized("read", type, obj, ctx))))
                return null

            /*  trace access  */
            yield (this.trace(obj.id, type, "create", "direct", "one"))

            /*  return new entity  */
            return obj
        }.bind(this))
    }

    /*  API: direct update a single entity  */
    mutationEntityUpdateOne (type) {
        return co.wrap(function * (parent, args, ctx, info) {
            /*  determine fields of entity as defined in GraphQL schema  */
            let defined = this.fieldsOfGraphQLType(info, type)

            /*  determine fields of entity as requested in GraphQL request  */
            let build = this.fieldsOfGraphQLRequest(args, info, type)

            /*  load entity instance  */
            let obj = yield (this.models[type].findById(args.id))
            if (obj === null)
                throw new Error(`entity ${type}#${args.id} not found`)

            /*  check access to entity  */
            if (!(yield (this.authorized("update", type, obj, ctx))))
                throw new Error(`not allowed to update entity of type "${type}"`)

            /*  adjust the attributes according to the request  */
            obj.update(build.attribute)

            /*  adjust the relationships according to the request  */
            yield (this.mutationEntityUpdateFields(type, obj,
                defined.relation, build.relation))

            /*  check access to entity again  */
            if (!(yield (this.authorized("read", type, obj, ctx))))
                return null

            /*  trace access  */
            yield (this.trace(obj.id, type, "update", "direct", "one"))

            /*  return updated entity  */
            return obj
        }.bind(this))
    }

    /*  API: direct update of multiple entities  */
    mutationEntityUpdateMany (type) {
        return co.wrap(function * (parent, args, ctx, info) {
            /*  determine fields of entity as defined in GraphQL schema  */
            let defined = this.fieldsOfGraphQLType(info, type)

            /*  determine fields of entity as requested in GraphQL request  */
            let build = this.fieldsOfGraphQLRequest(args, info, type)

            /*  load entity instances  */
            let opts = this.findAllOptions(type, args, info)
            let objs = yield (this.models[type].findAll(opts))
            if (objs.length === 0)
                throw new Error(`no such entities of type "${type}" found`)

            /*  iterate over all entities  */
            for (let i = 0; i < objs.length; i++) {
                let obj = objs[i]

                /*  check access to entity  */
                if (!(yield (this.authorized("update", type, obj, ctx))))
                    return new Error(`not allowed to update entity of type "${type}"`)

                /*  adjust the attributes according to the request  */
                obj.update(build.attribute)

                /*  adjust the relationships according to the request  */
                yield (this.mutationEntityUpdateFields(type, obj,
                    defined.relation, build.relation))
            }

            /*  check access to entities again  */
            objs = yield (Promise.filter(objs, (instance) => {
                return this.authorized("read", type, instance, ctx)
            }))

            /*  trace access  */
            yield (Promise.each(objs, (obj) => {
                return this.trace(obj.id, type, "update", "direct", "many")
            }))

            /*  return updated entites  */
            return objs
        }.bind(this))
    }

    /*  API: direct delete a single entity  */
    mutationEntityDeleteOne (type) {
        return co.wrap(function * (parent, args, ctx /*, info */) {
            /*  load entity instance  */
            let obj = yield (this.models[type].findById(args.id))
            if (obj === null)
                throw new Error(`entity ${type}#${args.id} not found`)

            /*  check access to target  */
            if (!(yield (this.authorized("delete", type, obj, ctx))))
                return new Error(`not allowed to delete entity of type "${type}"`)

            /*  delete the instance  */
            let result = obj.id
            yield (obj.destroy())

            /*  trace access  */
            yield (this.trace(result, type, "delete", "direct", "one"))

            /*  return id of deleted entity  */
            return result
        }.bind(this))
    }

    /*  API: direct delete multiple entities  */
    mutationEntityDeleteMany (type) {
        return co.wrap(function * (parent, args, ctx, info) {
            /*  load entity instances  */
            let opts = this.findAllOptions(type, args, info)
            let objs = yield (this.models[type].findAll(opts))
            if (objs.length === 0)
                throw new Error(`no such entities of type "${type}" found`)

            /*  iterate over all entities  */
            let result = []
            for (let i = 0; i < objs.length; i++) {
                let obj = objs[i]

                /*  check access to entity  */
                if (!(yield (this.authorized("delete", type, obj, ctx))))
                    throw new Error(`not allowed to delete entity "${type}#${obj.id}"`)

                /*  adjust the attributes according to the request  */
                result.push(obj.id)
                yield (obj.destroy())
            }

            /*  trace access  */
            yield (Promise.each(result, (id) => {
                return this.trace(id, type, "delete", "direct", "many")
            }))

            /*  return ids of deleted entities  */
            return result
        }.bind(this))
    }
}

